@@ -882,7 +882,7 @@ int bch2_bkey_durability(struct btree_trans *trans, struct bkey_s_c k, struct bk
 		if (d < 0)
 			return d;
 
-		if (test_bit(p.ptr.dev, c->devs_online.d))
+		if (bch2_dev_idx_is_online(c, p.ptr.dev))
 			ret->online += d;
 		ret->total += d;
 	}
@@ -1153,7 +1153,7 @@ bool bch2_bkey_devs_rw(struct bch_fs *c, struct bkey_s_c k)
 
 	guard(rcu)();
 	bkey_for_each_ptr(ptrs, ptr) {
-		CLASS(bch2_dev_tryget, ca)(c, ptr->dev);
+		struct bch_dev *ca = bch2_dev_rcu(c, ptr->dev);
 		if (!ca || ca->mi.state != BCH_MEMBER_STATE_rw)
 			return false;
 	}
@@ -1216,6 +1216,40 @@ bool bch2_bkey_matches_ptr(struct bch_fs *c, struct bkey_s_c k,
 	return false;
 }
 
+bool bch2_bkey_ptrs_match(struct bkey_s_c k1, struct extent_ptr_decoded p1,
+			  struct bkey_s_c k2, struct extent_ptr_decoded p2)
+{
+	return ((p1.ptr.dev		== p2.ptr.dev ||
+		 (p1.has_ec && p2.has_ec &&
+		  p1.ec.idx		== p2.ec.idx &&
+		  p1.ec.block		== p2.ec.block)) &&
+
+		p1.ptr.gen		== p2.ptr.gen &&
+
+		/*
+		 * This checks that the two pointers point
+		 * to the same region on disk - adjusting
+		 * for the difference in where the extents
+		 * start, since one may have been trimmed:
+		 */
+		(s64) p1.ptr.offset + p1.crc.offset - bkey_start_offset(k1.k) ==
+		(s64) p2.ptr.offset + p2.crc.offset - bkey_start_offset(k2.k) &&
+
+		/*
+		 * This additionally checks that the
+		 * extents overlap on disk, since the
+		 * previous check may trigger spuriously
+		 * when one extent is immediately partially
+		 * overwritten with another extent (so that
+		 * on disk they are adjacent) and
+		 * compression is in use:
+		 */
+		((p1.ptr.offset >= p2.ptr.offset &&
+		  p1.ptr.offset  < p2.ptr.offset + p2.crc.compressed_size) ||
+		 (p2.ptr.offset >= p1.ptr.offset &&
+		  p2.ptr.offset  < p1.ptr.offset + p1.crc.compressed_size)));
+}
+
 /*
  * Returns true if two extents refer to the same data:
  */
@@ -1235,31 +1269,7 @@ bool bch2_extents_match(const struct bch_fs *c, struct bkey_s_c k1, struct bkey_
 
 		bkey_for_each_ptr_decode(k1.k, ptrs1, p1, entry1)
 			bkey_for_each_ptr_decode(k2.k, ptrs2, p2, entry2)
-				if (p1.ptr.dev		== p2.ptr.dev &&
-				    p1.ptr.gen		== p2.ptr.gen &&
-
-				    /*
-				     * This checks that the two pointers point
-				     * to the same region on disk - adjusting
-				     * for the difference in where the extents
-				     * start, since one may have been trimmed:
-				     */
-				    (s64) p1.ptr.offset + p1.crc.offset - bkey_start_offset(k1.k) ==
-				    (s64) p2.ptr.offset + p2.crc.offset - bkey_start_offset(k2.k) &&
-
-				    /*
-				     * This additionally checks that the
-				     * extents overlap on disk, since the
-				     * previous check may trigger spuriously
-				     * when one extent is immediately partially
-				     * overwritten with another extent (so that
-				     * on disk they are adjacent) and
-				     * compression is in use:
-				     */
-				    ((p1.ptr.offset >= p2.ptr.offset &&
-				      p1.ptr.offset  < p2.ptr.offset + p2.crc.compressed_size) ||
-				     (p2.ptr.offset >= p1.ptr.offset &&
-				      p2.ptr.offset  < p1.ptr.offset + p1.crc.compressed_size)))
+				if (bch2_bkey_ptrs_match(k1, p1, k2, p2))
 					return true;
 
 		return false;
@@ -1371,6 +1381,29 @@ void bch2_extent_ptr_set_cached(struct bch_fs *c,
 	maybe_drop_cached_ptr(c, opts, k, ptr, have_cached_ptr);
 }
 
+static void bch2_bkey_set_ptrs_cached_mask(struct bch_fs *c,
+					   struct bch_inode_opts *opts,
+					   struct bkey_i *k,
+					   unsigned mask)
+{
+	while (mask) {
+		unsigned drop = __fls(mask);
+		mask &= ~BIT(drop);
+
+		bkey_for_each_ptr(bch2_bkey_ptrs(bkey_i_to_s(k)), ptr)
+			if (!drop--) {
+				bch2_extent_ptr_set_cached(c, opts, bkey_i_to_s(k), ptr);
+				break;
+			}
+	}
+}
+
+#if 0
+void bch2_bkey_set_ptrs_cached_mask()
+{
+}
+#endif
+
 static bool bch2_bkey_has_stale_ptrs(struct bch_fs *c, struct bkey_s_c k)
 {
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
@@ -1442,10 +1475,14 @@ void bch2_bkey_drop_extra_cached_ptrs(struct bch_fs *c,
 
 int bch2_bkey_drop_extra_durability(struct btree_trans *trans,
 				    struct bch_inode_opts *opts,
-				    struct bkey_s k)
+				    struct bkey_i *k,
+				    unsigned mask,
+				    bool kill)
 {
+	if (!mask)
+		return 0;
+
 	struct bch_fs *c = trans->c;
-	u8 ptrs_kill = 0;
 	u8 ptr_durability[BCH_BKEY_PTRS_MAX];
 	unsigned durability = 0, nr_ptrs = 0;
 
@@ -1454,39 +1491,74 @@ int bch2_bkey_drop_extra_durability(struct btree_trans *trans,
 	union bch_extent_entry *entry;
 	struct extent_ptr_decoded p;
 
-	bkey_for_each_ptr_decode(k.k, bch2_bkey_ptrs(k), p, entry) {
+	bkey_for_each_ptr_decode(&k->k, bch2_bkey_ptrs(bkey_i_to_s(k)), p, entry) {
 		BUG_ON(nr_ptrs >= ARRAY_SIZE(ptr_durability));
 
-		int d = bch2_extent_ptr_durability(trans, &p);
+		if (p.ptr.cached)
+			mask &= ~BIT(nr_ptrs);
+
+		int d = bch2_dev_idx_is_online(c, p.ptr.dev)
+			? bch2_extent_ptr_durability(trans, &p)
+			: 0;
 		if (d < 0)
 			return d;
 
 		BUG_ON(d > U8_MAX);
 
-		if (!d && !p.ptr.cached)
-			ptrs_kill |= BIT(nr_ptrs);
-
 		ptr_durability[nr_ptrs++] = d;
 		durability += d;
 	}
 
+	u8 ptrs_kill = 0;
 	for (unsigned i = 0; i < nr_ptrs; i++)
-		if (ptr_durability[i] &&
+		if ((mask & BIT(i)) &&
 		    durability - ptr_durability[i] >= opts->data_replicas) {
 			durability -= ptr_durability[i];
 			ptr_durability[i] = 0;
 			ptrs_kill |= BIT(i);
 		}
 
-	while (ptrs_kill) {
-		unsigned drop = __fls(ptrs_kill);
-		ptrs_kill &= ~BIT(drop);
+	if (ptrs_kill)
+		bch2_bkey_drop_ptrs_mask(c, k, ptrs_kill);
+	else
+		bch2_bkey_set_ptrs_cached_mask(c, opts, k, ptrs_kill);
+	return 0;
+}
 
-		bkey_for_each_ptr(bch2_bkey_ptrs(k), ptr)
-			if (!drop--) {
-				bch2_extent_ptr_set_cached(c, opts, k, ptr);
-				break;
+int bch2_bkey_drop_extra_ec_durability(struct btree_trans *trans,
+				       struct bch_inode_opts *opts,
+				       struct bkey_i *k, unsigned mask)
+{
+	if (!mask)
+		return 0;
+
+	struct bkey_durability durability;
+	try(bch2_bkey_durability(trans, bkey_i_to_s_c(k), &durability));
+
+	struct bch_fs *c = trans->c;
+	union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+	unsigned ptr_bit = 1;
+
+	bkey_for_each_ptr_decode(&k->k, bch2_bkey_ptrs(bkey_i_to_s(k)), p, entry) {
+		if (mask & ptr_bit) {
+			int d;
+			if (bch2_dev_idx_is_online(c, p.ptr.dev)) {
+				d = bch2_extent_ptr_durability(trans, &p);
+				if (d < 0)
+					return d;
+
+				d -= bch2_dev_durability(c, p.ptr.dev);
+			} else {
+				d = 0;
 			}
+
+			if (durability.online - d >= opts->data_replicas) {
+				durability.online -= d;
+				bch2_bkey_drop_ec(c, k, p.ptr.dev);
+			}
+		}
+		ptr_bit <<= 1;
 	}
 
 	return 0;
@@ -1673,6 +1745,7 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 	unsigned nr_ptrs = 0;
 	bool have_written = false, have_unwritten = false, have_ec = false, crc_since_last_ptr = false;
 	bool have_non_inval_dev_ptrs = false;
+	bool have_non_inval_dev_ptrs_dirty = false;
 	int ret = 0;
 
 	if (bkey_is_btree_ptr(k.k))
@@ -1705,12 +1778,14 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 			else
 				have_unwritten = true;
 
+			if (entry->ptr.dev != BCH_SB_MEMBER_INVALID || have_ec) {
+				if (!entry->ptr.cached)
+					have_non_inval_dev_ptrs_dirty = true;
+				have_non_inval_dev_ptrs = true;
+			}
+
 			have_ec = false;
 			crc_since_last_ptr = false;
-
-			if (entry->ptr.dev != BCH_SB_MEMBER_INVALID)
-				have_non_inval_dev_ptrs = true;
-
 			nr_ptrs++;
 			break;
 		case BCH_EXTENT_ENTRY_crc32:
@@ -1758,7 +1833,6 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 					 c, ptr_stripe_redundant,
 					 "redundant stripe entry");
 			have_ec = true;
-			have_non_inval_dev_ptrs = true;
 			break;
 		case BCH_EXTENT_ENTRY_reconcile:
 			try(bch2_extent_reconcile_validate(c, k, from, &entry->reconcile));
@@ -1792,14 +1866,20 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 			 c, extent_ptrs_redundant_stripe,
 			 "redundant stripe entry");
 
-	/*
-	 * we don't use KEY_TYPE_error for dead btree nodes - we still want the
-	 * other fields in bch_btree_ptr_v2
-	 */
-	bkey_fsck_err_on(!bkey_is_btree_ptr(k.k) &&
-			 !have_non_inval_dev_ptrs && !have_ec,
-			 c, extent_ptrs_all_invalid,
-			 "extent ptrs all to BCH_SB_MEMBER_INVALID");
+	if (!bkey_is_btree_ptr(k.k)) {
+		/*
+		 * we don't use KEY_TYPE_error for dead btree nodes - we still want the
+		 * other fields in bch_btree_ptr_v2
+		 */
+		bkey_fsck_err_on(!have_non_inval_dev_ptrs,
+				 c, extent_ptrs_all_invalid,
+				 "extent without valid pointers");
+
+		bkey_fsck_err_on(from.from == BKEY_VALIDATE_commit &&
+				 !have_non_inval_dev_ptrs_dirty,
+				 c, extent_ptrs_all_invalid,
+				 "extent without valid dirty pointers");
+	}
 fsck_err:
 	return ret;
 }
