@@ -36,6 +36,7 @@ void bch2_prt_csum_type(struct printbuf *,		enum bch_csum_type);
 void bch2_prt_compression_type(struct printbuf *,	enum bch_compression_type);
 void bch2_prt_str_hash_type(struct printbuf *,		enum bch_str_hash_type);
 void bch2_prt_reconcile_accounting_type(struct printbuf *, enum bch_reconcile_accounting_type);
+void bch2_prt_key_type_error_reason(struct printbuf *,	enum bch_key_type_errors);
 
 static inline const char *bch2_d_type_str(unsigned d_type)
 {
@@ -161,16 +162,6 @@ enum fsck_err_opts {
 	  OPT_UINT(1, BCH_REPLICAS_MAX + 1),				\
 	  BCH_SB_DATA_REPLICAS_WANT,	1,				\
 	  "#",		"Number of data replicas")			\
-	x(metadata_replicas_required, u8,				\
-	  OPT_FS|OPT_FORMAT|OPT_MOUNT,					\
-	  OPT_UINT(1, BCH_REPLICAS_MAX),				\
-	  BCH_SB_META_REPLICAS_REQ,	1,				\
-	  "#",		NULL)						\
-	x(data_replicas_required,	u8,				\
-	  OPT_FS|OPT_FORMAT|OPT_MOUNT,					\
-	  OPT_UINT(1, BCH_REPLICAS_MAX + 1),				\
-	  BCH_SB_DATA_REPLICAS_REQ,	1,				\
-	  "#",		NULL)						\
 	x(encoded_extent_max,		u32,				\
 	  OPT_FS|OPT_FORMAT|						\
 	  OPT_HUMAN_READABLE|OPT_MUST_BE_POW_2|OPT_SB_FIELD_SECTORS|OPT_SB_FIELD_ILOG2,\
@@ -568,6 +559,17 @@ enum fsck_err_opts {
 	  NULL,		"BTREE_ITER_prefetch causes btree nodes to be\n"\
 	  " prefetched sequentially")
 
+enum bch_opt_id {
+#define x(_name, ...)	Opt_##_name,
+	BCH_OPTS()
+#undef x
+	bch2_opts_nr
+};
+
+struct bch_opts_mask {
+	unsigned long	d[BITS_TO_LONGS(bch2_opts_nr)];
+};
+
 struct bch_opts {
 #define x(_name, _bits, ...)	unsigned _name##_defined:1;
 	BCH_OPTS()
@@ -605,13 +607,6 @@ static inline struct bch_opts bch2_opts_empty(void)
 
 void bch2_opts_apply(struct bch_opts *, struct bch_opts);
 
-enum bch_opt_id {
-#define x(_name, ...)	Opt_##_name,
-	BCH_OPTS()
-#undef x
-	bch2_opts_nr
-};
-
 struct bch_fs;
 struct printbuf;
 
@@ -662,6 +657,7 @@ void bch2_opt_to_text(struct printbuf *, struct bch_fs *, struct bch_sb *,
 void bch2_opts_to_text(struct printbuf *,
 		       struct bch_opts,
 		       struct bch_fs *, struct bch_sb *,
+		       struct bch_opts_mask *,
 		       unsigned, unsigned, unsigned);
 
 int bch2_opt_hook_pre_set(struct bch_fs *, struct bch_dev *, u64, enum bch_opt_id, u64, bool);
@@ -704,4 +700,11 @@ void bch2_inode_opts_get(struct bch_fs *, struct bch_inode_opts *, bool);
 bool bch2_opt_is_inode_opt(enum bch_opt_id);
 void bch2_inode_opts_to_text(struct printbuf *, struct bch_fs *, struct bch_inode_opts);
 
+void bch2_opt_change_unlock(struct bch_fs *);
+void bch2_opt_change_lock(struct bch_fs *);
+
+DEFINE_GUARD(opt_change_lock, struct bch_fs *,
+	     bch2_opt_change_lock(_T),
+	     bch2_opt_change_unlock(_T))
+
 #endif /* _BCACHEFS_OPTS_H */
