@@ -6,7 +6,6 @@
 #include "alloc/replicas.h"
 
 #include "data/checksum.h"
-#include "data/ec.h"
 #include "data/extents_sb.h"
 
 #include "journal/journal.h"
@@ -78,6 +77,7 @@ int bch2_set_version_incompat(struct bch_fs *c, enum bcachefs_metadata_version v
 {
 	if (((c->sb.features & BIT_ULL(BCH_FEATURE_incompat_version_field)) &&
 	     version <= c->sb.version_incompat_allowed)) {
+		guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 		guard(mutex)(&c->sb_lock);
 
 		if (version > c->sb.version_incompat) {
@@ -107,6 +107,37 @@ int bch2_set_version_incompat(struct bch_fs *c, enum bcachefs_metadata_version v
 	}
 }
 
+int bch2_set_version_incompat_locked(struct bch_fs *c, enum bcachefs_metadata_version version)
+{
+	if (((c->sb.features & BIT_ULL(BCH_FEATURE_incompat_version_field)) &&
+	     version <= c->sb.version_incompat_allowed)) {
+		if (version > c->sb.version_incompat) {
+			SET_BCH_SB_VERSION_INCOMPAT(c->disk_sb.sb,
+				max(BCH_SB_VERSION_INCOMPAT(c->disk_sb.sb), version));
+			bch2_write_super(c);
+		}
+		return 0;
+	} else {
+		BUILD_BUG_ON(BCH_VERSION_MAJOR(bcachefs_metadata_version_current) != 1);
+
+		unsigned minor = BCH_VERSION_MINOR(version);
+
+		if (!test_bit(minor, c->incompat_versions_requested) &&
+		    !test_and_set_bit(minor, c->incompat_versions_requested)) {
+			CLASS(printbuf, buf)();
+			prt_str(&buf, "requested incompat feature ");
+			bch2_version_to_text(&buf, version);
+			prt_str(&buf, " currently not enabled, allowed up to ");
+			bch2_version_to_text(&buf, c->sb.version_incompat_allowed);
+			prt_printf(&buf, "\n  set version_upgrade=incompatible to enable");
+
+			bch_notice(c, "%s", buf.buf);
+		}
+
+		return bch_err_throw(c, may_not_use_incompat_feature);
+	}
+}
+
 const char * const bch2_sb_fields[] = {
 #define x(name, nr)	#name,
 	BCH_SB_FIELDS()
@@ -186,7 +217,7 @@ void bch2_free_super(struct bch_sb_handle *sb)
 	kfree(sb->holder);
 	kfree(sb->sb_name);
 
-	kfree(sb->sb);
+	kvfree(sb->sb);
 	memset(sb, 0, sizeof(*sb));
 }
 
@@ -224,7 +255,7 @@ int bch2_sb_realloc(struct bch_sb_handle *sb, unsigned u64s)
 	if (dynamic_fault("bcachefs:add:super_realloc"))
 		return -BCH_ERR_ENOMEM_sb_realloc_injected;
 
-	new_sb = krealloc(sb->sb, new_buffer_size, GFP_NOFS|__GFP_ZERO);
+	new_sb = kvrealloc(sb->sb, new_buffer_size, GFP_NOFS|__GFP_ZERO);
 	if (!new_sb)
 		return -BCH_ERR_ENOMEM_sb_buf_realloc;
 
@@ -604,6 +635,9 @@ static void bch2_sb_update(struct bch_fs *c)
 	c->sb.clean		= BCH_SB_CLEAN(src);
 	c->sb.encryption_type	= BCH_SB_ENCRYPTION_TYPE(src);
 
+	c->sb.extent_bp_shift = BCH_SB_EXTENT_BP_SHIFT(c->disk_sb.sb) ?:
+		BCH_SB_EXTENT_BP_SHIFT_DEFAULT;
+
 	c->sb.nsec_per_time_unit = le32_to_cpu(src->time_precision);
 	c->sb.time_units_per_sec = NSEC_PER_SEC / c->sb.nsec_per_time_unit;
 
@@ -1186,12 +1220,12 @@ int bch2_write_super(struct bch_fs *c)
 	unsigned nr_members =	bch2_sb_nr_devices(c->disk_sb.sb);
 
 	if (!nr_wrote ||
-	    !bch2_can_read_fs_with_devs(c, sb_written, degraded_flags, NULL)) {
+	    !bch2_can_read_fs_with_devs(c, &sb_written, degraded_flags, NULL)) {
 		prt_printf(&err, "Unable to write superblock to sufficient devices (from %ps)\n",
 			   (void *) _RET_IP_);
 		prt_printf(&err, "Would not be able to mount with written devices\n");
 
-		bch2_can_read_fs_with_devs(c, sb_written, degraded_flags, &err);
+		bch2_can_read_fs_with_devs(c, &sb_written, degraded_flags, &err);
 
 		prt_printf(&err, "Wrote to %u/%u devices:\n", nr_wrote, nr_members);
 		scoped_guard(printbuf_indent, &err)
@@ -1201,7 +1235,7 @@ int bch2_write_super(struct bch_fs *c)
 		scoped_guard(printbuf_indent, &err)
 			bch2_devs_mask_to_text_locked(&err, c, &sb_unwritten);
 
-		if (bch2_fs_emergency_read_only2(c, &err))
+		if (bch2_fs_emergency_read_only(c, &err))
 			bch2_print_str(c, KERN_ERR, err.buf);
 	}
 out:
@@ -1215,6 +1249,7 @@ out:
 
 void __bch2_check_set_feature(struct bch_fs *c, unsigned feat)
 {
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 	if (!(c->sb.features & BIT_ULL(feat))) {
 		c->disk_sb.sb->features[0] |= cpu_to_le64(BIT_ULL(feat));
@@ -1268,6 +1303,7 @@ void bch2_sb_upgrade(struct bch_fs *c, unsigned new_version, bool incompat)
 
 void bch2_sb_upgrade_incompat(struct bch_fs *c)
 {
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 
 	if (c->sb.version == c->sb.version_incompat_allowed)
@@ -1289,6 +1325,8 @@ void bch2_sb_upgrade_incompat(struct bch_fs *c)
 
 	bch2_sb_set_upgrade_incompat(c, c->sb.version_incompat_allowed, c->sb.version);
 	bch2_write_super(c);
+
+	bch2_run_async_recovery_passes(c);
 }
 
 static int bch2_sb_ext_validate(struct bch_sb *sb, struct bch_sb_field *f,
@@ -1302,7 +1340,9 @@ static int bch2_sb_ext_validate(struct bch_sb *sb, struct bch_sb_field *f,
 	return 0;
 }
 
-static void bch2_sb_ext_to_text(struct printbuf *out, struct bch_sb *sb,
+static void bch2_sb_ext_to_text(struct printbuf *out,
+				struct bch_fs *c,
+				struct bch_sb *sb,
 				struct bch_sb_field *f)
 {
 	struct bch_sb_field_ext *e = field_to_type(f, ext);
@@ -1363,13 +1403,15 @@ static int bch2_sb_field_validate(struct bch_sb *sb, struct bch_sb_field *f,
 		prt_printf(err, "Invalid superblock section %s: %s",
 			   bch2_sb_fields[type], field_err.buf);
 		prt_newline(err);
-		bch2_sb_field_to_text(err, sb, f);
+		bch2_sb_field_to_text(err, NULL, sb, f);
 	}
 
 	return ret;
 }
 
-void __bch2_sb_field_to_text(struct printbuf *out, struct bch_sb *sb,
+void __bch2_sb_field_to_text(struct printbuf *out,
+			     struct bch_fs *c,
+			     struct bch_sb *sb,
 			     struct bch_sb_field *f)
 {
 	unsigned type = le32_to_cpu(f->type);
@@ -1379,10 +1421,12 @@ void __bch2_sb_field_to_text(struct printbuf *out, struct bch_sb *sb,
 		printbuf_tabstop_push(out, 32);
 
 	if (ops->to_text)
-		ops->to_text(out, sb, f);
+		ops->to_text(out, c, sb, f);
 }
 
-void bch2_sb_field_to_text(struct printbuf *out, struct bch_sb *sb,
+void bch2_sb_field_to_text(struct printbuf *out,
+			   struct bch_fs *c,
+			   struct bch_sb *sb,
 			   struct bch_sb_field *f)
 {
 	unsigned type = le32_to_cpu(f->type);
@@ -1395,13 +1439,11 @@ void bch2_sb_field_to_text(struct printbuf *out, struct bch_sb *sb,
 	prt_printf(out, " (size %zu):", vstruct_bytes(f));
 	prt_newline(out);
 
-	__bch2_sb_field_to_text(out, sb, f);
+	__bch2_sb_field_to_text(out, c, sb, f);
 }
 
 void bch2_sb_layout_to_text(struct printbuf *out, struct bch_sb_layout *l)
 {
-	unsigned i;
-
 	prt_printf(out, "Type:                    %u", l->layout_type);
 	prt_newline(out);
 
@@ -1413,7 +1455,7 @@ void bch2_sb_layout_to_text(struct printbuf *out, struct bch_sb_layout *l)
 	prt_newline(out);
 
 	prt_str(out, "Offsets:                 ");
-	for (i = 0; i < l->nr_superblocks; i++) {
+	for (unsigned i = 0; i < l->nr_superblocks; i++) {
 		if (i)
 			prt_str(out, ", ");
 		prt_printf(out, "%llu", le64_to_cpu(l->sb_offset[i]));
@@ -1421,7 +1463,8 @@ void bch2_sb_layout_to_text(struct printbuf *out, struct bch_sb_layout *l)
 	prt_newline(out);
 }
 
-void bch2_sb_to_text(struct printbuf *out, struct bch_sb *sb,
+void bch2_sb_to_text(struct printbuf *out,
+		     struct bch_fs *c, struct bch_sb *sb,
 		     bool print_layout, unsigned fields)
 {
 	if (!out->nr_tabstops)
@@ -1538,6 +1581,6 @@ void bch2_sb_to_text(struct printbuf *out, struct bch_sb *sb,
 	vstruct_for_each(sb, f)
 		if (fields & (1 << le32_to_cpu(f->type))) {
 			prt_newline(out);
-			bch2_sb_field_to_text(out, sb, f);
+			bch2_sb_field_to_text(out, c, sb, f);
 		}
 }
