@@ -16,28 +16,32 @@
 #include "init/passes.h"
 #include "init/progress.h"
 
-int bch2_dev_missing_bkey(struct bch_fs *c, struct bkey_s_c k, unsigned dev)
+int bch2_dev_missing_bkey_msg(struct bch_fs *c, struct bkey_s_c k, unsigned dev,
+			      struct printbuf *out)
 {
-	CLASS(printbuf, buf)();
-	bch2_log_msg_start(c, &buf);
+	if (dev == BCH_SB_MEMBER_INVALID)
+		return 0;
 
 	bool removed = test_bit(dev, c->devs_removed.d);
 
-	prt_printf(&buf, "pointer to %s device %u in key\n",
+	prt_printf(out, "pointer to %s device %u in key\n",
 		   removed ? "removed" : "nonexistent", dev);
-	bch2_bkey_val_to_text(&buf, c, k);
-	prt_newline(&buf);
+	bch2_bkey_val_to_text(out, c, k);
+	prt_newline(out);
 
-	bool print = removed
-		? bch2_count_fsck_err(c, ptr_to_removed_device, &buf)
-		: bch2_count_fsck_err(c, ptr_to_invalid_device, &buf);
+	if (removed)
+		bch2_count_fsck_err(c, ptr_to_removed_device, out);
+	else
+		bch2_count_fsck_err(c, ptr_to_invalid_device, out);
 
-	int ret = bch2_run_explicit_recovery_pass(c, &buf,
+	return bch2_run_explicit_recovery_pass(c, out,
 					BCH_RECOVERY_PASS_check_allocations, 0);
+}
 
-	if (print)
-		bch2_print_str(c, KERN_ERR, buf.buf);
-	return ret;
+int bch2_dev_missing_bkey(struct bch_fs *c, struct bkey_s_c k, unsigned dev)
+{
+	CLASS(bch_log_msg, msg)(c);
+	return bch2_dev_missing_bkey_msg(c, k, dev, &msg.m);
 }
 
 void bch2_dev_missing_atomic(struct bch_fs *c, unsigned dev)
@@ -337,6 +341,7 @@ void bch2_member_to_text_short(struct printbuf *out,
 			       struct bch_fs *c,
 			       struct bch_dev *ca)
 {
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 	bch2_member_to_text_short_locked(out, c, ca);
 }
@@ -497,8 +502,10 @@ void bch2_dev_io_errors_to_text(struct printbuf *out, struct bch_dev *ca)
 	struct bch_fs *c = ca->fs;
 	struct bch_member m;
 
-	scoped_guard(mutex, &ca->fs->sb_lock)
+	scoped_guard(memalloc_flags, PF_MEMALLOC_NOFS) {
+		guard(mutex)(&c->sb_lock);
 		m = bch2_sb_member_get(c->disk_sb.sb, ca->dev_idx);
+	}
 
 	printbuf_tabstop_push(out, 12);
 
@@ -524,6 +531,7 @@ void bch2_dev_errors_reset(struct bch_dev *ca)
 {
 	struct bch_fs *c = ca->fs;
 
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 
 	struct bch_member *m = bch2_members_v2_get_mut(c->disk_sb.sb, ca->dev_idx);
@@ -626,6 +634,7 @@ void bch2_dev_btree_bitmap_mark_locked(struct bch_fs *c, struct bkey_s_c k, bool
 
 void bch2_dev_btree_bitmap_mark(struct bch_fs *c, struct bkey_s_c k)
 {
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 	bool write_sb = false;
 	bch2_dev_btree_bitmap_mark_locked(c, k, &write_sb);
@@ -655,7 +664,8 @@ int bch2_btree_bitmap_gc(struct bch_fs *c)
 	struct progress_indicator progress;
 	bch2_progress_init(&progress, __func__, c, 0, ~0ULL);
 
-	scoped_guard(mutex, &c->sb_lock) {
+	scoped_guard(memalloc_flags, PF_MEMALLOC_NOFS) {
+		guard(mutex)(&c->sb_lock);
 		guard(rcu)();
 		for_each_member_device_rcu(c, ca, NULL)
 			ca->btree_allocated_bitmap_gc = 0;
@@ -680,7 +690,8 @@ int bch2_btree_bitmap_gc(struct bch_fs *c)
 
 	u64 sectors_marked_old = 0, sectors_marked_new = 0;
 
-	scoped_guard(mutex, &c->sb_lock) {
+	scoped_guard(memalloc_flags, PF_MEMALLOC_NOFS) {
+		guard(mutex)(&c->sb_lock);
 		struct bch_sb_field_members_v2 *mi = bch2_sb_field_get(c->disk_sb.sb, members_v2);
 
 		scoped_guard(rcu)
@@ -820,6 +831,7 @@ int bch2_sb_member_alloc(struct bch_fs *c)
 
 void bch2_sb_members_clean_deleted(struct bch_fs *c)
 {
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 	bool write_sb = false;
 
@@ -850,6 +862,8 @@ void __bch2_dev_mi_field_upgrades(struct bch_fs *c, struct bch_dev *ca, bool *wr
 void bch2_dev_mi_field_upgrades(struct bch_dev *ca)
 {
 	struct bch_fs *c = ca->fs;
+
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 	bool write_sb = false;
 
@@ -864,6 +878,7 @@ void bch2_dev_mi_field_upgrades(struct bch_dev *ca)
  */
 void bch2_fs_mi_field_upgrades(struct bch_fs *c)
 {
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 	bool write_sb = false;
 
