@@ -17,6 +17,52 @@
 
 #include <linux/random.h>
 
+/*
+ * Snapshot trees:
+ *
+ * A node in a snapshot tree references keys with that snapshot ID, and all keys
+ * with ancestor snapshot IDs not overwritten by a descendent snapshot.
+ *
+ * When a subvolume is deleted, we now have dead and redundant snapshot nodes
+ * that must be cleaned up.
+ *
+ * - Dead:
+ *
+ *   A snapshot node with no children, and without a subvolume pointing to it,
+ *   is unreferenced and can be deleted
+ *
+ * - Redundant:
+ *
+ *   Interior snapshot nodes (nodes with children) are only referenced by their
+ *   child snapshot nodes. An interior node with only one child is redundant; we
+ *   can clean it up by moving all non-overwritten keys to the child snapshot
+ *   and removing it from the snapshot tree.
+ *
+ * Snapshot node states:
+ *
+ * - WILL_DELETE: this doesn't need to be a separate state bit. Indicates a leaf
+ *   node that's no longer referenced by a subvolume (bch_snapshot.subvol == 0),
+ *   so it's pending deletion
+ *
+ * - NO_KEYS: We can't remove interior nodes from the snapshot tree at runtime,
+ *   because that can require changing bch_snapshot.depth on arbitrarily many
+ *   children, and we can't do that atomically.
+ *
+ *   So instead, at runtime we'll do the heavy lifting of removing all keys that
+ *   reference that snapshot ID, leave it in a half dead state, and the next
+ *   time we start up we'll remove it from the snapshot tree.
+ *
+ *   Technically, we could, because the codepaths where this matters use the
+ *   RCU-protected snapshot table - but there's a lot of work that has to be
+ *   done for deleting interior snapshot nodes; parent/child pointers need to be
+ *   updated, skiplists need to be adjusted, and if we get any of this wrong
+ *   things can and will go horrifically wrong.
+ *
+ *   But if we defer it until recovery, when we're not yet running multithreaded,
+ *   we can also run the check_snapshots recovery pass afterwards, for extra
+ *   safety.
+ */
+
 static void bch2_snapshot_delete_nodes_to_text(struct printbuf *out, struct snapshot_delete *d, bool full)
 {
 	size_t limit = !full ? 10 : SIZE_MAX;
@@ -266,7 +312,14 @@ static int delete_dead_snapshots_process_key(struct btree_trans *trans,
 					     struct btree_iter *iter,
 					     struct bkey_s_c k)
 {
-	struct snapshot_delete *d = &trans->c->snapshots.delete;
+	struct bch_fs *c = trans->c;
+	struct snapshot_delete *d = &c->snapshots.delete;
+
+	int ret = bch2_check_key_has_snapshot(trans, iter, k);
+	if (ret < 0)
+		return ret;
+	if (ret)
+		return bch2_trans_commit_lazy(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc);
 
 	if (snapshot_list_has_id(&d->delete_leaves, k.k->p.snapshot))
 		return bch2_btree_delete_at(trans, iter,
@@ -274,6 +327,8 @@ static int delete_dead_snapshots_process_key(struct btree_trans *trans,
 
 	u32 live_child = interior_delete_has_id(&d->delete_interior, k.k->p.snapshot);
 	if (live_child) {
+		BUG_ON(!bch2_snapshot_exists(c, live_child));
+
 		struct bkey_i *new = errptr_try(bch2_bkey_make_mut_noupdate(trans, k));
 
 		new->k.p.snapshot = live_child;
@@ -459,16 +514,21 @@ static int check_should_delete_snapshot(struct btree_trans *trans, struct bkey_s
 		u32 id = le32_to_cpu(s.v->children[i]);
 		if (id && !snapshot_list_has_id(&d->delete_leaves, id)) {
 			nr_live_children++;
-			live_child = interior_delete_has_id(&d->delete_interior, id) ?: id;
+
+			live_child = interior_delete_has_id(&d->delete_interior, id) ?:
+				interior_delete_has_id(&d->no_keys, id) ?:
+				id;
+
+			BUG_ON(live_child && !bch2_snapshot_exists(c, live_child));
 		}
 	}
 
-	if (nr_live_children == 2 ||
-	    (nr_live_children == 1 && BCH_SNAPSHOT_NO_KEYS(s.v)))
+	if (nr_live_children == 2)
 		return 0;
 
-	try(snapshot_list_add_nodup(c, &d->deleting_from_trees,
-				    bch2_snapshot_tree(c, s.k->p.offset)));
+	if (!BCH_SNAPSHOT_NO_KEYS(s.v))
+		try(snapshot_list_add_nodup(c, &d->deleting_from_trees,
+					    bch2_snapshot_tree(c, s.k->p.offset)));
 
 	if (!nr_live_children) {
 		try(snapshot_list_add(c, &d->delete_leaves, s.k->p.offset));
@@ -478,10 +538,18 @@ static int check_should_delete_snapshot(struct btree_trans *trans, struct bkey_s
 			.live_child	= live_child,
 		};
 
-		if (n.id == n.live_child) {
-			bch_err(c, "error finding live descendent of %llu", s.k->p.offset);
-			return -EINVAL;
-		}
+		BUG_ON(n.id == n.live_child);
+		BUG_ON(!bch2_snapshot_is_ancestor(c, n.live_child, n.id));
+
+		/*
+		 * We're not doing any processing for NO_KEYS snapshot nodes,
+		 * but we still track them so that we can find the correct
+		 * live_child when deleting parents, above:
+		 */
+		if (!BCH_SNAPSHOT_NO_KEYS(s.v))
+			try(darray_push(&d->delete_interior, n));
+		else
+			try(darray_push(&d->no_keys, n));
 	}
 
 	return 0;
@@ -512,9 +580,6 @@ static int bch2_fix_child_of_deleted_snapshot(struct btree_trans *trans,
 	struct bch_fs *c = trans->c;
 	u32 nr_deleted_ancestors = 0;
 
-	if (!bch2_snapshot_exists(c, k.k->p.offset))
-		return 0;
-
 	if (k.k->type != KEY_TYPE_snapshot)
 		return 0;
 
@@ -614,6 +679,7 @@ int __bch2_delete_dead_snapshots(struct bch_fs *c)
 
 	scoped_guard(mutex, &d->progress_lock) {
 		darray_exit(&d->deleting_from_trees);
+		darray_exit(&d->no_keys);
 		darray_exit(&d->delete_interior);
 		darray_exit(&d->delete_leaves);
 		d->running = false;
@@ -704,11 +770,13 @@ int bch2_delete_dead_interior_snapshots(struct bch_fs *c)
 		{
 			CLASS(bch_log_msg_level, msg)(c, LOGLEVEL_notice);
 
-			prt_printf(&msg.m, "Deleting interior snapshot nodes forces check_snapshots:");
+			prt_printf(&msg.m, "Deleting interior snapshot nodes forces check_snapshots:\n");
 			try(bch2_run_explicit_recovery_pass(c, &msg.m,
 					BCH_RECOVERY_PASS_check_snapshots, 0));
 		}
 
+		try(bch2_check_snapshots_trans(trans));
+
 		/*
 		 * Fixing children of deleted snapshots can't be done completely
 		 * atomically, if we crash between here and when we delete the interior
